/**
 * TMNode.java Version 32l 5-Feb-1999
 * Copyright (c) 1997-2000 Kyushu University
 * 
 * Updated 99-09-08
 *
 * Bug fixed 99.08.10
 *   mkSchain: wrong skip-chain was created for
 *             -> p(f(g(a))),p(f(k(b))),p(f(g(c))).
 *
 * updated 8-August-1999
 *    based on mm-mgtp0.5d/TMNode.java (updated v32g's on 3-May-1999)
 *    to deal with the case where a symbol idetntical with
 *    a constant symbol is generated.
 *    e.g., ..{M=2}->a(M). ->a(2).  
 *
 * arity count method (revised version of cmgtp32i)
 *   duplication of bind occurs for p(f(X?,Y?))
 *                              but not for p(X?,Y?)
 * modified to skip chain method:
 *      based on cmgtp32j (arity count method)
 * Use a new Term.java and MGTPParser.jj to support that
 * FVar has a maxst field pointing to its max-subterm.
 **/

package mgtp;

class TMNode {
  public static int PREDS,FUNCS;
  public static TMNode NIL=new TMNode();
  public static MGTP mgtp;
  public static boolean createOn;
  public static boolean top_maxst;

  TMNode up,downL,downR,left,right;
  TMNode lftlf,rgtlf;
  ACell ndac;
  ACell gen;
  TMNode[] farray;
  Term uterm;
  Term sterm; // ATMNode,TMLeaf: max_sterm || FTMNode: null
  boolean gensym;

  TMNode() {}

  TMNode(String s, MGTP mg) { // Top node
    //Counter.index++;
    up=downL=left=right=NIL; downR=this; ndac=ACell.OFF;
    farray=new TMNode[PREDS];
    uterm=new ATermL("Top");
    mgtp=mg;
    gen=mgtp.ac;
  }

  public String toString() {return toString(0);}
  String toString(int i) {
    if (this==NIL) return "";
    String s="";

    for (int j=0; j<i; j++) s+=" ";
    s+="$$ "+(uterm.name==null ? ""+((ITerm)uterm).val : uterm.name)+
    "'s down: ";
    for (int k=0; k<PREDS; k++) {
      TMNode n=farray[k];
      s+=(n==null ? "" : ""+n.uterm+", ");
    }
    s+="\n";

    for (int j=0; j<i; j++) s+=" ";
    s+=(uterm.name==null ? ""+((ITerm)uterm).val : uterm.name);
    if (downL!=NIL) s+="\n"+downL.toString(i+1);
    if (right!=NIL) s+="\n"+right.toString(i);
    return s;
  }

  int leafCount() {
    int c=0;
    if (this==NIL) return 0;
    if (downL!=NIL) c+=downL.leafCount();
    if (right!=NIL) c+=right.leafCount();
    return c;
  }

  TMNode findNode(Term t) {
    if (downL==NIL) return null;
    TMNode n=downL;
    if (t instanceof ITerm) {
      int i=((ITerm)t).val;
      while (n!=NIL) {
	if (((ITerm)n.uterm).val==i) {
	  // %%%
	  if (t.index!=0) {
	    farray[t.index]=n; 
	    n.uterm=t; // %%% added
	  }
	  // %%%
	  return n;
	}
	n=n.right;  
      }
      return null;  
    }
    else {
      while (n!=NIL) {
	if (n.uterm.name==t.name) return n;
	n=n.right;  
      }
      return null;  
    }
  }

  /*******  ixBV/ixBVL ***************************/

  TMNode ixBV(Term t) {
    boolean done=false;
    Term tt; TMNode n,l;
    //Counter.index++;
    tt=((t.index== -2) ? t.arg.arg : t);
    n=farray[tt.index];
    if (n!=null || ((gensym || tt.index==0) 
		    && (n=findNode(tt))!=null)) {
                         // My Node exists
      if (tt.arg==null) { // I am AT
	if (t.next==null) // ST Leaf
	  return n;
	else 
	  return n.ixBV(t.next); // Through the value from t.next
      } 
      else { // I am Func
	if (!top_maxst) top_maxst=done=true;

	if (t.next==null) { // CTL : without next
	  l=n.ixBV(tt.arg);
	  if (createOn) {
	    ((FTMNode)n).mkSchain(l,mgtp.ac);

	    if (t instanceof CCTerm) ((CCTerm)t).anode=l;
	    else l.sterm=tt;
	  }
	  return l;
	} 
	else { // CT : with next
	  l=n.ixBV(tt.arg);
	  if (createOn) {
	    ((FTMNode)n).mkSchain(l,mgtp.ac);

	    if (t instanceof CCTerm) ((CCTerm)t).anode=l;
	    else l.sterm=tt;
	  }

	  if (done) top_maxst=false;   // essential!
	  return l.ixBV(t.next); // Through the value from t.next
	}
      }
    } 
    createOn=true;
    if (downR!=NIL) return crDRBV(t);
    else return crDBV(t);
  }

  TMNode ixBVL(Term t) {
    boolean done=false;
    Term tt; TMNode n,l;
    //Counter.index++;
    tt=((t.index== -2) ? t.arg.arg : t);
    n=farray[tt.index];

    if (n!=null || ((gensym || tt.index==0) 
    		    && (n=findNode(tt))!=null)) {
                         // My Node exists
      if (tt.arg==null) { // I am AT
	if (t.next==null) // Leaf
	  return n;
	else 
	  return n.ixBVL(t.next); // Through the value from t.next
      } 
      else { // I am Func
	if (!top_maxst) top_maxst=done=true;
	
	if (t.next==null) { // CTL : without next
	  l=n.ixBVL(tt.arg);
	  if (createOn) {
	    ((FTMNode)n).mkSchain(l,mgtp.ac);

	    if (t instanceof CCTerm) ((CCTerm)t).anode=l;
	    else l.sterm=tt;
	  }
	  return l;
	} 
	else { // CT : with next
	  l=n.ixBV(tt.arg);
	  if (createOn) {
	    ((FTMNode)n).mkSchain(l,mgtp.ac);

	    if (t instanceof CCTerm) ((CCTerm)t).anode=l;
	    else l.sterm=tt;
	  }

	  if (done) top_maxst=false;   // essential!
	  return l.ixBVL(t.next); // Through the value from t.next
	}
      }
    }
    createOn=true;
    if (downR!=NIL) return crDRBVL(t);
    else return crDBVL(t);
  }

  /*******  crDBV/crDBVL/crDRBV/crDRBVL ***************************/

  TMNode crDBV(Term t) {
    boolean done=false; 
    Term tt; TMNode n,l;
    //Counter.index++;
    tt=((t.index== -2) ? t.arg.arg : t);

    if (tt.arg==null) { // I am AT
      if (tt.index==0) {
	n=downR=downL=new ATMNode(tt,this); gensym=true;}
      else n=farray[tt.index]=downR=downL=new ATMNode(tt,this);
      if (t.next==null) // ST Leaf
	return n;
      else 
	return n.crDBV(t.next); // Through the value from t.next
    } 
    else { // I am Func
      n=farray[tt.index]=downR=downL=new FTMNode(tt,this);
      if (!top_maxst) ((FTMNode)n).maxst=top_maxst=done=true;

      if (t.next==null) { // CTL : without next
	l=n.crDBV(tt.arg);
	((FTMNode)n).mkSchain(l,mgtp.ac);

	if (t instanceof CCTerm) ((CCTerm)t).anode=l;
	else l.sterm=tt;
	return l;
      } 
      else { // CT : with next
	((FTMNode)n).maxst=true;

	l=n.crDBV(tt.arg);
	((FTMNode)n).mkSchain(l,mgtp.ac);

	if (t instanceof CCTerm) ((CCTerm)t).anode=l;
	else l.sterm=tt;

	if (done) done=top_maxst=false;
	return l.crDBV(t.next); // Through the value from t.next
      }
    }
  }

  TMNode crDBVL(Term t) {
    boolean done=false; 
    Term tt; TMNode n,l;
    //Counter.index++;
    tt=((t.index== -2) ? t.arg.arg : t);

    if (tt.arg==null) { // I am AT
      if (t.next==null) { // Leaf
	if (tt.index==0) {
	  n=downR=downL=new TMLeaf(tt,this); gensym=true;}
	else n=farray[tt.index]=downR=downL=new TMLeaf(tt,this);
	return n;
      }
      else {
	if (tt.index==0) {
	  n=downR=downL=new ATMNode(tt,this); gensym=true;}
	else n=farray[tt.index]=downR=downL=new ATMNode(tt,this);
	return n.crDBVL(t.next); // Through the value from t.next
      } 
    }
    else { // I am Func
      n=farray[tt.index]=downR=downL=new FTMNode(tt,this);
      if (!top_maxst) ((FTMNode)n).maxst=top_maxst=done=true;

      if (t.next==null) { // CTL : without next
	l=n.crDBVL(tt.arg);
	((FTMNode)n).mkSchain(l,mgtp.ac);

	if (t instanceof CCTerm) ((CCTerm)t).anode=l;
	else l.sterm=tt;
	return l;
      } 
      else { // CT : with next

	((FTMNode)n).maxst=true;

	l=n.crDBV(tt.arg);
	((FTMNode)n).mkSchain(l,mgtp.ac);

	if (t instanceof CCTerm) ((CCTerm)t).anode=l;
	else l.sterm=tt;

	if (done) done=top_maxst=false;
	return l.crDBVL(t.next); // Through the value from t.next
      }
    }
  }

  TMNode crDRBV(Term t) {
    boolean done=false; 
    Term tt; TMNode n,l;
    //Counter.index++;
    tt=((t.index== -2) ? t.arg.arg : t);

    if (tt.arg==null) { // I am AT
      if (downR.gen!=mgtp.ac) {
	mgtp.rnl=new RNdList(downR,mgtp.rnl);
      }
      if (tt.index==0) {
	n=downR=downR.right=new ATMNode(tt,this,downR); gensym=true;
      }
      else n=farray[tt.index]=downR=downR.right=new ATMNode(tt,this,downR);
      if (t.next==null) // ST Leaf
	return n;
      else 
	return n.crDBV(t.next);
    } 
    else { // I am Func
      if (downR.gen!=mgtp.ac) {
	mgtp.rnl=new RNdList(downR,mgtp.rnl);
      }
      n=farray[tt.index]=downR=downR.right=new FTMNode(tt,this,downR);
      if (!top_maxst) ((FTMNode)n).maxst=top_maxst=done=true;

      if (t.next==null) { // CTL : without next
	l=n.crDBV(tt.arg);
	((FTMNode)n).mkSchain(l,mgtp.ac);

	if (t instanceof CCTerm) ((CCTerm)t).anode=l;
	else l.sterm=tt;
	return l;
      } 
      else { // CT : with next

	((FTMNode)n).maxst=true;

	l=n.crDBV(tt.arg);
	((FTMNode)n).mkSchain(l,mgtp.ac);

	if (t instanceof CCTerm) ((CCTerm)t).anode=l;
	else l.sterm=tt;

	if (done) done=top_maxst=false;
	return l.crDBV(t.next);
      }
    }
  }

  TMNode crDRBVL(Term t) {
    boolean done=false; 
    Term tt; TMNode n,l;
    //Counter.index++;
    tt=((t.index== -2) ? t.arg.arg : t);

    if (tt.arg==null) { // I am AT
      if (t.next==null) { // Leaf
	if (downR.gen!=mgtp.ac) {
	  mgtp.rnl=new RNdList(downR,mgtp.rnl);
	}
	if (tt.index==0) {
	  n=downR=downR.right=new TMLeaf(tt,this,downR); gensym=true;
	}
	else 
	  n=farray[tt.index]=downR=downR.right=new TMLeaf(tt,this,downR);
	return n;
      }
      else {
	if (downR.gen!=mgtp.ac) {
	  mgtp.rnl=new RNdList(downR,mgtp.rnl);
	}
	if (tt.index==0) {
	  n=downR=downR.right=new ATMNode(tt,this,downR); gensym=true;
	}
	else n=farray[tt.index]=downR=downR.right=new ATMNode(tt,this,downR);
	return n.crDBVL(t.next);
      } 
    }
    else { // I am Func
      if (downR.gen!=mgtp.ac) {
	mgtp.rnl=new RNdList(downR,mgtp.rnl);
      }
      n=farray[tt.index]=downR=downR.right=new FTMNode(tt,this,downR);
      if (!top_maxst) ((FTMNode)n).maxst=top_maxst=done=true;

      if (t.next==null) { // CTL : without next
	l=n.crDBVL(tt.arg);
	((FTMNode)n).mkSchain(l,mgtp.ac);

	if (t instanceof CCTerm) ((CCTerm)t).anode=l;
	else l.sterm=tt;
	return l;
      } 
      else { // CT : with next

	((FTMNode)n).maxst=true;

	l=n.crDBV(tt.arg);
	((FTMNode)n).mkSchain(l,mgtp.ac);

	if (t instanceof CCTerm) ((CCTerm)t).anode=l;
	else l.sterm=tt;

	if (done) done=top_maxst=false;
	return l.crDBVL(t.next);
      }
    }
  }

  /**********************************/

  void matchTMBV(Term t, AConj ai) {}
  void skipTM(Term t, AConj ai) {}
  TMNode fmatch(Term t) {return NIL;}
  TMNode fmatchL(Term t) {return NIL;}
  void mkSchain(TMNode lfn, ACell ac) {}

}

class TMLeaf extends TMNode {
  GTerm term,plit,nlit;

  TMLeaf(Term t, TMNode u) {
    //Counter.index++;
    up=u; downL=downR=left=right=NIL; ndac=ACell.OFF;
    uterm=t;
    gen=mgtp.ac;
    lftlf=rgtlf=NIL;
  }

  TMLeaf(Term t, TMNode u, TMNode l) {
    //Counter.index++;
    up=u; left=l; downL=downR=right=NIL; ndac=ACell.OFF;
    uterm=t;
    gen=mgtp.ac;
    lftlf=rgtlf=NIL;
  }

  public String toString(int i) {
    String s="";
    for (int j=0; j<i; j++) s+=" ";
    return s+=(uterm.name==null ? ""+((ITerm)uterm).val : 
	       uterm.name)+": "+
      " --> sterm: "+sterm+" "+
      " lftlf: "+(lftlf!=NIL ? lftlf.uterm.name : "NIL")+" "+
      " rghlf: "+(rgtlf!=NIL ? rgtlf.uterm.name : "NIL")+"\n"+
      " plit: "+plit+" nlit: "+nlit+"\n"+right.toString(i);
  }

  int leafCount() {
    int c=1;
    if (right!=NIL) c+=right.leafCount();
    return c;
  }

  TMNode fmatchL(Term t) {
    //Counter.matchTM++;
    if (ndac.act) return this; else return NIL;
  }

  void skipTM(Term t, AConj ai) {
    //Counter.matchTM++;
    t.arg=uterm;  // bind X?

    ai.hd.leafProc(term,ai);
  }

  void matchTMBV(Term t, AConj ai) {
    ai.hd.leafProc(term,ai);
  }

}

class ATMNode extends TMNode {

  ATMNode(Term t, TMNode u) {
    //Counter.index++;
    up=u; downL=downR=left=right=NIL; ndac=ACell.OFF;
    farray=new TMNode[FUNCS];
    uterm=t;
    gen=mgtp.ac;
    lftlf=rgtlf=NIL;
  }

  ATMNode(Term t, TMNode u, TMNode l) {
    //Counter.index++;
    up=u; left=l; downL=downR=right=NIL; ndac=ACell.OFF;
    farray=new TMNode[FUNCS];
    uterm=t;
    gen=mgtp.ac;
    lftlf=rgtlf=NIL;
  }

  public String toString() {return toString(0);}


  String toString(int i) {
    if (this==NIL) return "";
    String s="";

    for (int j=0; j<i; j++) s+=" ";
    s+=(uterm.name==null ? "## "+((ITerm)uterm).val : uterm.name)+"'s down: ";
    for (int k=0; k<FUNCS; k++) {
      TMNode n=farray[k];
      s+=(n==null ? "" : ""+n.uterm+", ");
    }
    s+="\n";

    for (int j=0; j<i; j++) s+=" ";
    s+=(uterm.name==null ? ""+((ITerm)uterm).val : uterm.name);
    s+=" --> sterm: "+sterm;
    s+=" lftlf: "+(lftlf!=NIL ? lftlf.uterm.name : "NIL")+" ";
    s+=" rghlf: "+(rgtlf!=NIL ? rgtlf.uterm.name : "NIL");
    if (downL!=NIL) s+="\n"+downL.toString(i+1);
    if (right!=NIL) s+="\n"+right.toString(i);
    return s;
  }

  void matchTMBV(Term t, AConj ai) {
    TMNode n; Term ts=t.next; Term tt;
    //Counter.matchTM++;
    if (ts.index== -1) {
      n=downR;
      while (n!=NIL) {
	//Counter.nodeFV++;
	if (n.ndac.act) {
	  //Counter.nodeFVact++;
	  n.skipTM(ts,ai);
	  if (ai.mgtp.closed) return;
	}
	n=n.left;
      }
      return;
    }
    tt=((ts.index== -2) ? ts.arg.arg : ts);
    n=farray[tt.index];
    if (n==null && (gensym || tt.index==0)) n=findNode(tt);
    if (n!=null&&n.ndac.act) {
      n.matchTMBV(ts,ai); // 99.09.08
    }
  }

  TMNode fmatch(Term t) {
    TMNode n; Term ts=t.next;
    //Counter.matchTM++;
    if (ts==null) return this;
    n=farray[ts.index];
    if (n==null && (gensym || ts.index==0)) n=findNode(ts);
    if (n!=null&&n.ndac.act) return n.fmatch(ts);
    return NIL;
  }

  TMNode fmatchL(Term t) {
    TMNode n; Term ts=t.next;
    //Counter.matchTM++;
    if (ts==null) return this;
    n=farray[ts.index];
    if (n==null && (gensym || ts.index==0)) n=findNode(ts);
    if (n!=null&&n.ndac.act) return n.fmatchL(ts);
    return NIL;
  }

  void skipTM(Term t, AConj aj) {
                                      // matchTMBV unfolded
    TMNode n; Term ts=t.next; Term tt;
    //Counter.matchTM++;

    t.arg=uterm;  // bind X?

    // begin matchTMBV
    if (ts.index== -1) {
      n=downR;
      while (n!=NIL) {
	//Counter.nodeFV++;
	if (n.ndac.act) {
	  //Counter.nodeFVact++;
	  n.skipTM(ts,aj);
	  if (aj.mgtp.closed) return;
	}
	n=n.left;
      }
      return;
    }
    tt=((ts.index== -2) ? ts.arg.arg : ts);
    n=farray[tt.index];
    if (n==null && (gensym || tt.index==0)) n=findNode(tt);
    if (n!=null&&n.ndac.act) {
      n.matchTMBV(ts,aj);
    } // end of matchTMBV
  }

}


class PTMNode extends TMNode {

  PTMNode(Term t, TMNode u, TMNode l) {
    //Counter.index++;
    up=u; left=l; right=downL=downR=NIL; ndac=ACell.OFF;
    farray=new TMNode[FUNCS];
    uterm=t;
    gen=mgtp.ac;
  }

  String toString(int i) {
    if (this==NIL) return "";
    String s="";

    for (int j=0; j<i; j++) s+=" ";
    s+="&& "+(uterm.name==null ? ""+((ITerm)uterm).val : uterm.name)+
    "'s down: ";
    for (int k=0; k<FUNCS; k++) {
      TMNode n=farray[k];
      s+=(n==null ? "" : ""+n.uterm+", ");
    }
    s+="\n";

    for (int j=0; j<i; j++) s+=" ";
    s+=(uterm.name==null ? ""+((ITerm)uterm).val : uterm.name);
    if (downL!=NIL) s+="\n"+downL.toString(i+1);
    if (right!=NIL) s+="\n"+right.toString(i);
    return s;
  }

  void matchTMBV(Term t, AConj aj) {
    TMNode n; Term ts=t.arg; Term tt;

    //Counter.matchTM++;
    if (ts.index== -1) {
      n=downR;
      while (n!=NIL) {
	//Counter.nodeFV++;
	if (n.ndac.act) {
	  //Counter.nodeFVact++;
	  n.skipTM(ts,aj);
	  if (aj.mgtp.closed) return;
	}
	n=n.left;
      }
      return;
    }
    tt=((ts.index== -2) ? ts.arg.arg : ts);
    n=farray[tt.index];
    if (n==null && (gensym || tt.index==0)) n=findNode(tt);
    if (n!=null&&n.ndac.act) {
      n.matchTMBV(ts,aj);
    }
  }

  void skipTM(Term t, AConj aj) {
                                    // for X,-X -> .. : not used
    //Counter.matchTM++;

    t.arg=uterm;  // bind X?

    matchTMBV(t,aj);
  }

}

class FTMNode extends TMNode {
  boolean maxst;

  FTMNode(Term t, TMNode u) {
    //Counter.index++;
    up=u; downL=downR=left=right=NIL; ndac=ACell.OFF;
    farray=new TMNode[FUNCS];
    uterm=t;
    gen=mgtp.ac;
    lftlf=rgtlf=NIL;
  }

  FTMNode(Term t, TMNode u, TMNode l) {
    //Counter.index++;
    up=u; left=l; downL=downR=right=NIL; ndac=ACell.OFF;
    farray=new TMNode[FUNCS];
    uterm=t;
    gen=mgtp.ac;
    lftlf=rgtlf=NIL;
  }

  public String toString() {return toString(0);}

  String toString(int i) {
    if (this==NIL) return "";
    String s="";

    for (int j=0; j<i; j++) s+=" ";
    s+=">> "+(uterm.name==null ? ""+((ITerm)uterm).val : uterm.name)+
    "'s down: ";
    for (int k=0; k<FUNCS; k++) {
      TMNode n=farray[k];
      s+=(n==null ? "" : ""+n.uterm+", ");
    }
    s+="\n";

    for (int j=0; j<i; j++) s+=" ";
    s+=(uterm.name==null ? ""+((ITerm)uterm).val : uterm.name);
    s+=" maxst: "+maxst;
    s+=" --> sterm: "+sterm;
    s+=" lftlf: "+(lftlf!=NIL ? lftlf.uterm.name : "NIL")+" ";
    s+=" rghlf: "+(rgtlf!=NIL ? rgtlf.uterm.name : "NIL");
    if (downL!=NIL) s+="\n"+downL.toString(i+1);
    if (right!=NIL) s+="\n"+right.toString(i);
    return s;
  }

  void mkSchain(TMNode lfn, ACell ac) {
    if (lftlf==NIL) {
      if (rgtlf!=NIL&&rgtlf.gen!=ac) {
	mgtp.rsc=new RSkChain(this,rgtlf,1,mgtp.rsc);
      }
      lftlf=rgtlf=lfn;
    }
    else {
      if (lfn.lftlf==NIL) {
	lfn.lftlf=rgtlf;         //                 <--  c
	TMNode rrlf=rgtlf.rgtlf;
	lfn.rgtlf=rrlf;          //                      c  -->
	if (rrlf!=NIL) {
	  if (rrlf.lftlf.gen!=ac) {
	    mgtp.rsc=new RSkChain(rrlf,rrlf.lftlf,0,mgtp.rsc);
	  }
	  rrlf.lftlf=lfn;        //                      c <==  b
	}
	if (rgtlf.gen!=ac) {
	  mgtp.rsc=new RSkChain(rgtlf,rrlf,1,mgtp.rsc);
	}
	rgtlf.rgtlf=lfn;                //               a  ==> c
	if (rgtlf!=NIL&&rgtlf.gen!=ac) {
	  mgtp.rsc=new RSkChain(this,rgtlf,1,mgtp.rsc);
	}
	//                                               f ==+
	//                                                   |
	rgtlf=lfn;               //                      c <=+
      } else {
	if (rgtlf==lfn.lftlf) {
	  if (rgtlf!=NIL&&rgtlf.gen!=ac) {
	    mgtp.rsc=new RSkChain(this,rgtlf,1,mgtp.rsc);
	  }
	  rgtlf=lfn;
	}
      }
    }
  }

  void matchTMBV(Term t, AConj aj) {

    TMNode n,l; Term ts,tt,tts; 
    //Counter.matchTM++;
    if (t.index== -2) { 
      tts=t.arg.arg.arg;
      n=farray[tts.index];
      if (n==null && (gensym || tts.index==0)) n=findNode(tts);
      if ((n!=null)&&n.ndac.act) {
	l=n.fmatchL(tts);
	if (l!=NIL) {
	  l.matchTMBV(t,aj);
	}
      }
    } 
    else {
      ts=t.arg;
      if (ts.index== -1) {
	n=downR;
	while (n!=NIL) {
	  //Counter.nodeFV++;
	  if (n.ndac.act) {
	    //Counter.nodeFVact++;
	    n.skipTM(ts,aj);
            if (aj.mgtp.closed) return;
	  }
	  n=n.left;
	}
	return;
      }
      tt=((ts.index== -2) ? ts.arg.arg : ts);
      n=farray[tt.index];
      if (n==null && (gensym || tt.index==0)) n=findNode(tt);
      if (n!=null&&n.ndac.act) {
	n.matchTMBV(ts,aj);
      }
    }
  }

  TMNode fmatch(Term t) {
    TMNode n; Term ts=t.arg;
    //Counter.matchTM++;
    n=farray[ts.index];
    if (n==null && (gensym || ts.index==0)) n=findNode(ts);
    if (n!=null&&n.ndac.act) {
      if (t.next==null)
	return n.fmatch(ts);
      else return n.fmatch(ts).fmatch(t);
    }
    return NIL;
  }

  TMNode fmatchL(Term t) {
    TMNode n; Term ts=t.arg;
    //Counter.matchTM++;
    n=farray[ts.index];
    if (n==null && (gensym || ts.index==0)) n=findNode(ts);
    if (n!=null&&n.ndac.act) {
      if (t.next==null)
	return n.fmatchL(ts);
      else return n.fmatch(ts).fmatchL(t);
    }
    return NIL;
  }

  void skipTM(Term t, AConj ai) {
                                    // t=?X,... 
    //Counter.matchTM++;

    TMNode l=rgtlf; 
    TMNode e=lftlf.lftlf;
    // Note that lfn may be ATMNode or TMLeaf

    while(l!=e) {
      if (l.ndac.act) {
	if (maxst) {((FVar)t).arg=l.sterm;}
	else ((FVar)t).maxst.bind(l.sterm,(FVar)t);
	l.matchTMBV(t,ai);
      }
      l=l.lftlf;
    }
  }
  
}


class RNdList {
  public static RNdList NIL=new RNdList();
  TMNode target; RNdList tl;

  RNdList() {}
  RNdList(TMNode n, RNdList t) {target=n; tl=t;} // Counter.rnl++;}

  public String toString() {
    RNdList r=this; String s="";
    while (r!=NIL) {
      s+="Node: "+r.target+"\n";
      r=r.tl;
    }
    return s;
  }

  void restore() {
    RNdList r=this;
    while (r!=NIL) {
      TMNode parn=r.target.up;
      TMNode curn=r.target.right;
      while (curn!=TMNode.NIL) {
	parn.farray[curn.uterm.index]=null;
	curn=curn.right;
      }
      parn.downR=r.target;
      r.target.right=TMNode.NIL;
      r=r.tl;
    }
  }

}


class RSkChain {
  public static RSkChain NIL=new RSkChain();
  TMNode target, prlf; RSkChain tl;
  int lr;  // lr=0 -> left, lr=1 -> right

  RSkChain() {}
  RSkChain(TMNode n, TMNode l, int dir, RSkChain t) 
  {target=n; prlf=l; lr=dir; tl=t;} // Counter.rsc++;}

  public String toString() {
    RSkChain r=this; String s="";
    while (r!=NIL) {
      s+="Node: "+r.target+"\n";
      r=r.tl;
    }
    return s;
  }

  void restore() {
    RSkChain r=this;
    while (r!=NIL) {
      if (r.lr==0) r.target.lftlf=r.prlf;
      else r.target.rgtlf=r.prlf;
      r=r.tl;
    }
    //    return;
  }

}


// eof
